@@ -33,9 +33,9 @@
 static void device_add_usage(void)
 {
 	puts("bcachefs device add - add a device to an existing filesystem\n"
-	     "Usage: bcachefs device add [OPTION]... filesystem device\n"
+	     "Usage: bcachefs device add [OPTION]... <filesystem> <device>\n"
 	     "\n"
-	     "Options:\n");
+	     "Options:");
 
 	bch2_opts_usage(OPT_FORMAT|OPT_DEVICE);
 
@@ -44,6 +44,7 @@ static void device_add_usage(void)
 	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
+	exit(EXIT_SUCCESS);
 }
 
 static int cmd_device_add(int argc, char *argv[])
@@ -135,9 +136,8 @@ static int cmd_device_add(int argc, char *argv[])
 
 static void device_remove_usage(void)
 {
-	puts("bcachefs device_remove - remove a device from a filesystem\n"
-	     "Usage:\n"
-	     "  bcachefs device remove <device>|<devid> <path>\n"
+	puts("bcachefs device remove - remove a device from a filesystem\n"
+	     "Usage: bcachefs device remove <device|dev_id> <path>\n"
 	     "\n"
 	     "Options:\n"
 	     "  -f, --force                  Force removal, even if some data couldn't be migrated\n"
@@ -205,13 +205,14 @@ static int cmd_device_remove(int argc, char *argv[])
 
 static void device_online_usage(void)
 {
-	puts("bcachefs device online - readd a device to a running filesystem\n"
-	     "Usage: bcachefs device online [OPTION]... device\n"
+	puts("bcachefs device online - re-add a device to a running filesystem\n"
+	     "Usage: bcachefs device online [OPTION]... <device>\n"
 	     "\n"
 	     "Options:\n"
 	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
+	exit(EXIT_SUCCESS);
 }
 
 static int cmd_device_online(int argc, char *argv[])
@@ -246,13 +247,14 @@ static int cmd_device_online(int argc, char *argv[])
 static void device_offline_usage(void)
 {
 	puts("bcachefs device offline - take a device offline, without removing it\n"
-	     "Usage: bcachefs device offline [OPTION]... device\n"
+	     "Usage: bcachefs device offline [OPTION]... <device>\n"
 	     "\n"
 	     "Options:\n"
 	     "  -f, --force                  Force, if data redundancy will be degraded\n"
 	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
+	exit(EXIT_SUCCESS);
 }
 
 static int cmd_device_offline(int argc, char *argv[])
@@ -292,29 +294,46 @@ static int cmd_device_offline(int argc, char *argv[])
 static void device_evacuate_usage(void)
 {
 	puts("bcachefs device evacuate - move data off of a given device\n"
-	     "Usage: bcachefs device evacuate [OPTION]... device\n"
+	     "Usage: bcachefs device evacuate [OPTION]... <device>\n"
 	     "\n"
 	     "Options:\n"
-	     "  -f, --force                  Force if data redundancy will be degraded\n"
 	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
+	exit(EXIT_SUCCESS);
+}
+
+static int evacuate_v0(struct bchfs_handle fs, unsigned dev_idx, const char *dev_path)
+{
+	struct bch_ioctl_dev_usage_v2 *u = bchu_dev_usage(fs, dev_idx);
+
+	if (u->state == BCH_MEMBER_STATE_rw) {
+		printf("Setting %s readonly\n", dev_path);
+		bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_ro, BCH_FORCE_IF_DEGRADED);
+	}
+
+	free(u);
+
+	return bchu_data(fs, (struct bch_ioctl_data) {
+		.op		= BCH_DATA_OP_migrate,
+		.start_btree	= 0,
+		.start_pos	= POS_MIN,
+		.end_btree	= BTREE_ID_NR,
+		.end_pos	= POS_MAX,
+		.migrate.dev	= dev_idx,
+	});
 }
 
 static int cmd_device_evacuate(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
-		{ "force",			no_argument,	NULL, 'f' },
 		{ "help",			no_argument,	NULL, 'h' },
 		{ NULL }
 	};
-	int opt, flags = 0;
+	int opt;
 
 	while ((opt = getopt_long(argc, argv, "fh", longopts, NULL)) != -1)
 		switch (opt) {
-		case 'f':
-			flags |= BCH_FORCE_IF_DEGRADED;
-			break;
 		case 'h':
 			device_evacuate_usage();
 			exit(EXIT_SUCCESS);
@@ -331,32 +350,41 @@ static int cmd_device_evacuate(int argc, char *argv[])
 	int dev_idx;
 	struct bchfs_handle fs = bchu_fs_open_by_dev(dev_path, &dev_idx);
 
-	struct bch_ioctl_dev_usage_v2 *u = bchu_dev_usage(fs, dev_idx);
+	if (bcachefs_kernel_version() < bcachefs_metadata_version_reconcile)
+		return evacuate_v0(fs, dev_idx, dev_path);
 
-	if (u->state == BCH_MEMBER_STATE_rw) {
-		printf("Setting %s readonly\n", dev_path);
-		bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_ro, flags);
+	printf("Setting %s evacuating \n", dev_path);
+	bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_evacuating, BCH_FORCE_IF_DEGRADED);
+
+	while (true) {
+		struct bch_ioctl_dev_usage_v2 *u = bchu_dev_usage(fs, dev_idx);
+
+		u64 data = 0;
+		for (unsigned type = 0; type < u->nr_data_types; type++)
+			if (!data_type_is_empty(type) &&
+			    !data_type_is_hidden(type))
+				data += u->d[type].sectors;
+		free(u);
+
+		printf("\33[2K\r");
+		CLASS(printbuf, buf)();
+		prt_units_u64(&buf, data << 9);
+
+		fputs(buf.buf, stdout);
+		fflush(stdout);
+
+		if (!data)
+			return 0;
+
+		sleep(1);
 	}
-
-	free(u);
-
-	return bchu_data(fs, (struct bch_ioctl_data) {
-		.op		= BCH_DATA_OP_migrate,
-		.start_btree	= 0,
-		.start_pos	= POS_MIN,
-		.end_btree	= BTREE_ID_NR,
-		.end_pos	= POS_MAX,
-		.migrate.dev	= dev_idx,
-	});
 }
 
 static void device_set_state_usage(void)
 {
-	puts("bcachefs device set-state\n"
-	     "Usage: bcachefs device set-state <new-state> <device>|<devid> <path>\n"
-	     "\n"
-	     "<new-state>: one of rw, ro, failed or spare\n"
-	     "<path>: path to mounted filesystem, optional unless specifying device by id\n"
+	puts("bcachefs device set-state - change the state of a device\n"
+	     "Usage: bcachefs device set-state <rw|ro|evacuating|spare> <device> [path]\n"
+	     "       bcachefs device set-state <rw|ro|evacuating|spare> <dev_id> <path>\n"
 	     "\n"
 	     "Options:\n"
 	     "  -f, --force                  Force if data redundancy will be degraded\n"
@@ -446,7 +474,7 @@ static int cmd_device_set_state(int argc, char *argv[])
 
 			bch2_write_super(c);
 		}
-		bch2_fs_stop(c);
+		bch2_fs_exit(c);
 		return ret;
 	}
 
@@ -473,8 +501,8 @@ static int cmd_device_set_state(int argc, char *argv[])
 
 static void device_resize_usage(void)
 {
-	puts("bcachefs device resize \n"
-	     "Usage: bcachefs device resize device [ size ]\n"
+	puts("bcachefs device resize - resize the filesystem on a device\n"
+	     "Usage: bcachefs device resize <device> <size>\n"
 	     "\n"
 	     "Options:\n"
 	     "  -h, --help                   Display this help and exit\n"
@@ -574,15 +602,15 @@ static int cmd_device_resize(int argc, char *argv[])
 			fprintf(stderr, "resize error: %s\n%s", bch2_err_str(ret), err.buf);
 
 		enumerated_ref_put(&resize->io_ref[READ], 0);
-		bch2_fs_stop(c);
+		bch2_fs_exit(c);
 	}
 	return 0;
 }
 
 static void device_resize_journal_usage(void)
 {
-	puts("bcachefs device resize-journal \n"
-	     "Usage: bcachefs device resize-journal device size\n"
+	puts("bcachefs device resize-journal - resize the journal on a device\n"
+	     "Usage: bcachefs device resize-journal <device> <size>\n"
 	     "\n"
 	     "Options:\n"
 	     "  -h, --help                   Display this help and exit\n"
@@ -673,7 +701,7 @@ static int cmd_device_resize_journal(int argc, char *argv[])
 			fprintf(stderr, "resize error: %s\n", bch2_err_str(ret));
 
 		enumerated_ref_put(&resize->io_ref[READ], 0);
-		bch2_fs_stop(c);
+		bch2_fs_exit(c);
 	}
 	return 0;
 }
@@ -681,7 +709,7 @@ static int cmd_device_resize_journal(int argc, char *argv[])
 static int device_usage(void)
 {
        puts("bcachefs device - manage devices within a running filesystem\n"
-            "Usage: bcachefs device <CMD> [OPTION]\n"
+            "Usage: bcachefs device <CMD> [OPTION]...\n"
             "\n"
             "Commands:\n"
             "  add                          Add a new device to an existing filesystem\n"
@@ -689,12 +717,12 @@ static int device_usage(void)
             "  online                       Re-add an existing member to a filesystem\n"
             "  offline                      Take a device offline, without removing it\n"
             "  evacuate                     Migrate data off a specific device\n"
-            "  set-state                    Mark a device as failed\n"
+            "  set-state                    Change device state (rw, ro, evacuating, spare)\n"
             "  resize                       Resize filesystem on a device\n"
             "  resize-journal               Resize journal on a device\n"
             "\n"
             "Report bugs to <linux-bcachefs@vger.kernel.org>");
-       return 0;
+       exit(EXIT_SUCCESS);
 }
 
 int device_cmds(int argc, char *argv[])
