@@ -1376,11 +1376,8 @@ int bch2_journal_read(struct bch_fs *c, struct journal_start_info *info)
 			set_bit(JOURNAL_degraded, &c->journal.flags);
 	}
 
-	if (!sysctl_hung_task_timeout_secs)
-		closure_sync(&jlist.cl);
-	else
-		while (closure_sync_timeout(&jlist.cl, sysctl_hung_task_timeout_secs * HZ / 2))
-			;
+	while (closure_sync_timeout(&jlist.cl, HZ))
+		;
 
 	if (jlist.ret)
 		return jlist.ret;
