@@ -4,6 +4,7 @@ mod dump_stack;
 mod logging;
 mod wrappers;
 mod device_scan;
+mod http;
 
 use std::{
     ffi::{c_char, CString},
@@ -112,6 +113,12 @@ fn main() -> ExitCode {
         None => args[1].as_str(),
     };
 
+    // fuse will call this after daemonizing, we can't create threads before - note that mount
+    // may invoke fusemount, via -t bcachefs.fuse
+    if cmd != "mount" && cmd != "fusemount" {
+        unsafe { c::linux_shrinkers_init() };
+    }
+
     match cmd {
         "completions" => {
             commands::completions(args[1..].to_vec());
